# type: ignore
import ast
import asyncio
import os
from datetime import datetime, timezone
from typing import List, Optional

import click

from uipath._cli._evals._runtime import UiPathEvalContext, UiPathEvalRuntime
from uipath._cli._runtime._contracts import (
    UiPathRuntimeContext,
    UiPathRuntimeContextBuilder,
    UiPathRuntimeFactory,
)
from uipath._cli._runtime._runtime import UiPathRuntime
from uipath._cli.middlewares import MiddlewareResult, Middlewares

from .._utils.constants import ENV_JOB_ID
from ..telemetry import track
from ._utils._console import ConsoleLogger

console = ConsoleLogger()


class LiteralOption(click.Option):
    def type_cast_value(self, ctx, value):
        try:
            return ast.literal_eval(value)
        except Exception as e:
            raise click.BadParameter(value) from e


def eval_agent_middleware(
    entrypoint: Optional[str] = None,
    eval_set: Optional[str] = None,
    eval_ids: Optional[List[str]] = None,
    workers: int = 8,
    no_report: bool = False,
    **kwargs,
) -> MiddlewareResult:
    def generate_eval_context(
        runtime_context: UiPathRuntimeContext,
    ) -> UiPathEvalContext:
        os.makedirs("evals/results", exist_ok=True)
        timestamp = datetime.now(timezone.utc).strftime("%M-%H-%d-%m-%Y")
        base_context = UiPathRuntimeContextBuilder().with_defaults().build()
        # TODO: the name should include the eval_set name. those files should not be commited to SW
        base_context.execution_output_file = (
            f"evals/results/{timestamp}.json"
            if not os.getenv("UIPATH_JOB_KEY")
            else None
        )
        return UiPathEvalContext(
            runtime_context=runtime_context,
            no_report=no_report,
            workers=workers,
            eval_set=eval_set,
            eval_ids=eval_ids,
            **kwargs,
            **base_context.model_dump(),
        )

    try:
        runtime_factory = UiPathRuntimeFactory(UiPathRuntime, UiPathRuntimeContext)
        context = (
            UiPathRuntimeContextBuilder()
            .with_defaults(**kwargs)
            .with_entrypoint(entrypoint)
            .with_entrypoint(entrypoint)
            .mark_eval_run()
            .build()
        )

        async def execute():
            async with UiPathEvalRuntime.from__eval_context(
                factory=runtime_factory, context=generate_eval_context(context)
            ) as eval_runtime:
                await eval_runtime.execute()

        asyncio.run(execute())
        return MiddlewareResult(should_continue=False)

    except Exception as e:
        return MiddlewareResult(
            should_continue=False, error_message=f"Error running evaluation: {str(e)}"
        )


@click.command()
@click.argument("entrypoint", required=False)
@click.argument("eval_set", required=False)
@click.option("--eval-ids", cls=LiteralOption, default="[]")
@click.option(
    "--no-report",
    is_flag=True,
    help="Do not report the evaluation results",
    default=False,
)
@click.option(
    "--workers",
    type=int,
    default=8,
    help="Number of parallel workers for running evaluations (default: 8)",
)
@track(when=lambda *_a, **_kw: os.getenv(ENV_JOB_ID) is None)
def eval(
    entrypoint: Optional[str],
    eval_set: Optional[str],
    eval_ids: List[str],
    no_report: bool,
    workers: int,
) -> None:
    """Run an evaluation set against the agent.

    Args:
        entrypoint: Path to the agent script to evaluate (optional, will auto-discover if not specified)
        eval_set: Path to the evaluation set JSON file (optional, will auto-discover if not specified)
        eval_ids: Optional list of evaluation IDs
        workers: Number of parallel workers for running evaluations
        no_report: Do not report the evaluation results
    """
    result = Middlewares.next(
        "eval",
        entrypoint,
        eval_set,
        eval_ids,
        no_report=no_report,
        workers=workers,
    )

    if result.should_continue:
        result = eval_agent_middleware(
            entrypoint=entrypoint,
            eval_set=eval_set,
            eval_ids=eval_ids,
            workers=workers,
            no_report=no_report,
        )
    if result.should_continue:
        console.error("Could not process the request with any available handler.")
    if result.error_message:
        console.error(result.error_message)

    console.success("Evaluation completed successfully")


if __name__ == "__main__":
    eval()



import copy
from collections import defaultdict
from time import time
from typing import Dict, Generic, List, Optional, Sequence, TypeVar

from opentelemetry.sdk.trace import ReadableSpan
from opentelemetry.sdk.trace.export import SpanExporter, SpanExportResult

from uipath.eval._helpers import auto_discover_entrypoint

from .._runtime._contracts import (
    UiPathBaseRuntime,
    UiPathRuntimeContext,
    UiPathRuntimeFactory,
    UiPathRuntimeResult,
    UiPathRuntimeStatus,
)
from .._utils._eval_set import EvalHelpers
from ._models import EvaluationItem
from ._models._agent_execution_output import UiPathEvalRunExecutionOutput

T = TypeVar("T", bound=UiPathBaseRuntime)
C = TypeVar("C", bound=UiPathRuntimeContext)


class ExecutionSpanExporter(SpanExporter):
    """Custom exporter that stores spans grouped by execution ids."""

    def __init__(self):
        # { execution_id -> list of spans }
        self._spans: Dict[str, List[ReadableSpan]] = defaultdict(list)

    def export(self, spans: Sequence[ReadableSpan]) -> SpanExportResult:
        for span in spans:
            if span.attributes is not None:
                exec_id = span.attributes.get("execution.id")
                if exec_id is not None and isinstance(exec_id, str):
                    self._spans[exec_id].append(span)

        return SpanExportResult.SUCCESS

    def get_spans(self, execution_id: str) -> List[ReadableSpan]:
        """Retrieve spans for a given execution id."""
        return self._spans.get(execution_id, [])

    def clear(self, execution_id: Optional[str] = None) -> None:
        """Clear stored spans for one or all executions."""
        if execution_id:
            self._spans.pop(execution_id, None)
        else:
            self._spans.clear()

    def shutdown(self) -> None:
        self.clear()


class UiPathEvalContext(UiPathRuntimeContext, Generic[C]):
    """Context used for evaluation runs."""

    runtime_context: C
    no_report: bool
    workers: int
    eval_set: Optional[str] = None
    eval_ids: Optional[List[str]] = None

    def __init__(
        self,
        runtime_context: C,
        no_report: bool,
        workers: int,
        eval_set: Optional[str] = None,
        eval_ids: Optional[List[str]] = None,
        **kwargs,
    ):
        super().__init__(
            runtime_context=runtime_context,  # type: ignore
            no_report=no_report,
            workers=workers,
            eval_set=eval_set,
            eval_ids=eval_ids,
            **kwargs,
        )
        self._auto_discover()

    def _auto_discover(self):
        self.runtime_context.entrypoint = (
            self.runtime_context.entrypoint or auto_discover_entrypoint()
        )
        self.eval_set = self.eval_set or EvalHelpers.auto_discover_eval_set()


class UiPathEvalRuntime(UiPathBaseRuntime, Generic[T, C]):
    """Specialized runtime for evaluation runs, with access to the factory."""

    def __init__(
        self, context: "UiPathEvalContext[C]", factory: "UiPathRuntimeFactory[T, C]"
    ):
        super().__init__(context)
        self.context: "UiPathEvalContext[C]" = context
        self.factory: UiPathRuntimeFactory[T, C] = factory
        self.span_exporter: ExecutionSpanExporter = ExecutionSpanExporter()
        self.factory.add_span_exporter(self.span_exporter)

    @classmethod
    def from__eval_context(
        cls,
        context: "UiPathEvalContext[C]",
        factory: "UiPathRuntimeFactory[T, C]",
    ) -> "UiPathEvalRuntime[T, C]":
        return cls(context, factory)

    async def execute(self) -> Optional[UiPathRuntimeResult]:
        """Evaluation logic. Can spawn other runtimes through the factory."""
        if self.context.eval_set is None:
            raise ValueError("eval_set must be provided for evaluation runs")

        evaluation_set = EvalHelpers.load_eval_set(
            self.context.eval_set, self.context.eval_ids
        )
        execution_output_list: list[UiPathEvalRunExecutionOutput] = []
        for eval_item in evaluation_set.evaluations:
            execution_output_list.append(await self.execute_agent(eval_item))

        self.context.result = UiPathRuntimeResult(
            output={
                "results": execution_output_list,
            },
            status=UiPathRuntimeStatus.SUCCESSFUL,
            resume=None,
        )

        return self.context.runtime_context.result

    def _prepare_new_runtime_context(self, eval_item: EvaluationItem) -> C:
        runtime_context = copy.deepcopy(self.context.runtime_context)
        runtime_context.execution_id = eval_item.id
        runtime_context.input_json = eval_item.inputs
        # here we can pass other values from eval_item: expectedAgentBehavior, simulationInstructions etc.
        return runtime_context

    # TODO: this would most likely need to be ported to a public AgentEvaluator class
    async def execute_agent(
        self, eval_item: EvaluationItem
    ) -> "UiPathEvalRunExecutionOutput":
        runtime_context = self._prepare_new_runtime_context(eval_item)
        start_time = time()
        result = await self.factory.execute_in_root_span(
            runtime_context, root_span=eval_item.name
        )
        end_time = time()
        if runtime_context.execution_id is None:
            raise ValueError("execution_id must be set for eval runs")

        spans = self.span_exporter.get_spans(runtime_context.execution_id)
        self.span_exporter.clear(runtime_context.execution_id)

        if result is None:
            raise ValueError("Execution result cannot be None for eval runs")

        return UiPathEvalRunExecutionOutput(
            execution_time=end_time - start_time,
            spans=spans,
            result=result,
        )

    async def cleanup(self) -> None:
        """Cleanup runtime resources."""
        pass

    async def validate(self) -> None:
        """Cleanup runtime resources."""
        pass
